class BaseLayer:
    def __init__(self):
        self.x = None
        self.y = None

    def forward_prop(self, x):
        raise NotImplementedError
    
    def backward_prop(self, dE_dY, learning_rate):
        raise NotImplementedError


import numpy

class Dense():
    def __init__(self, input_shape, output_shape):
        # Randomly initialize weights for each connection between input and output (-0.5 to 0.5)
        self.weights = numpy.random.rand(input_shape, output_shape) - 0.5 

        #Initialize a 0 for each output neuron as a bias
        self.bias = numpy.zeros((1, output_shape))

    def forward_prop(self, x):
        #Save input for backpropagation
        self.x = x

        #Return output based on given input by doing the dot product between the input and weights plus the bias
        self.y = numpy.dot(self.x, self.weights) + self.bias
        return self.y

    def backward_prop(self, dE_dY, learning_rate):
        #Calculate ∂E/∂W --> ∂E/∂W = ∂Y/∂W * ∂E/∂Y
        dE_dW = numpy.dot(self.x.T, dE_dY) #Note: T is transpose (flips shape of matrix)

        #Calculate ∂E/∂X --> ∂E/∂X = ∂E/∂Y * ∂Y/∂X
        dE_dX = numpy.dot(dE_dY, self.weights.T)

        dE_dB = dE_dY

        #Gradient Descent: Update weights and biases in the opposite direction of the gradient proportionately using learning rate
        self.bias -= dE_dB * learning_rate
        self.weights -= dE_dW * learning_rate

        #Return ∂E/∂X so it can be used as ∂E/∂Y for previous layer
        return dE_dX



import numpy

def sigmoid(x):
    return 1 / (1 + numpy.exp(-x))

def sigmoid_derivative(x):
    return sigmoid(x) * (1 - sigmoid(x))

def relu(x):
    return numpy.maximum(0,x)

def relu_derivative(x):
    copyX = x
    copyX[copyX < 0] = 0
    copyX[copyX > 0] = 1
    return x 

def tanh(x):
    return numpy.tanh(x);

def tanh_prime(x):
    return 1-numpy.tanh(x)**2;


class Model():
    def __init__(self):
        self.layers = []

    def add(self, layer):
        self.layers.append(layer)

    def predict(self, input):
        #Stores list of all outputs for each input
        output = []

        for i in range(len(input)):
            #Keeps track of current output for the ith input (By Default the first is the input layer)
            singleOutput = input[i]

            #Forward propogate each layer using the output of the last layer
            for layer in self.layers:
                singleOutput = layer.forward_prop(singleOutput)
            
            #Add to the output list
            output.append(singleOutput)

        return output
    
    def print_error(self, epoch, error):
        print(f"Loss of Epoch #{epoch}: {error}")

    def fit(self, x, y, loss, loss_derivative, epochs=1000, learning_rate=0.01, verbosity=1):
        x = [[i] for i in x]
        y = [[i] for i in y]

        for i in range(epochs):
            #Keeps track of error for current epoch for potential display at the end (depends on level of verbosity)
            total_error = 0

            for j in range(len(x)):
                #Keeps track of current output (starts with the input layer)
                output = x[j]

                #Forward propogate each layer using the output of the last layer
                for layer in self.layers:
                    output = layer.forward_prop(output)

                #Update the total error
                total_error += loss(y[j], output)

                #Backpropogation: Update weights and biases for each layer using derivative of error with respect to the output
                dE_dY = loss_derivative(y[j], output)
                #Loop backwards over layers as ∂E/∂X of the last layer is ∂E/∂Y of the layer before it
                for layer in reversed(self.layers):
                    dE_dY = layer.backward_prop(dE_dY, learning_rate)

            average_error = total_error / len(x)

            if verbosity == 0:
                continue
            elif verbosity == 1:
                if (i+1) % 100 == 0:
                    self.print_error(i+1, average_error)
            elif verbosity == 2:
                if (i+1) % 10 == 0:
                    self.print_error(i+1, average_error)
            else:
                self.print_error(i+1, average_error)


class ActivationLayer():
    def __init__(self, activation_function, activation_derivative):
        self.activation_function = activation_function
        self.activation_derivative = activation_derivative

    def forward_prop(self, x):
        self.x = x
        #pass input into activation function
        self.y = self.activation_function(self.x)
        return self.y
    
    def backward_prop(self, dE_dY, learning_rate):
        #Return multiplication of activation derivative and ∂E/∂Y as there is not parameters to optimize in this layer
        return self.activation_derivative(self.x) * dE_dY
        


from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='custom_neural_net_creator',
    version='1.2',
    license='MIT',
    author="Yogesh Seenichamy",
    author_email='yogeshseeni60@gmail.com',
    description="A Neural Network Module to create Custom Dense Neural Networks",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages('src'),
    package_dir={'': 'src'},
    url='https://github.com/YogeshSeeni/NeuralNetworkModule',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'numpy',
    ],
)


import numpy

def mean_squared_error(actual, prediction):
    return numpy.mean(numpy.power(actual - prediction, 2))

def mean_squared_error_derivative(actual, prediction):
    return 2/actual.size*(prediction-actual)